# -*- coding: iso-8859-15 -*-
"""pyauthd_nss.py

Implementation of the NSS plugin for the python authentication daemon. The NSS
plugin handles all communication with PyAuthD_NSS_*.c, the PyAuthD NSS modules.

Copyright (C) 2005, Heiko Wundram.
Released under a BSD-style license, see LICENSE for details.
"""

# Versioning.
__author__ = "$LastChangedBy: modelnine $"
__version__ = "$LastChangedRevision: 103 $"
__date__ = "$LastChangedDate: 2005-10-28 19:56:08 +0200 (pią, 28 paź 2005) $"

# Python imports.

import errno

# Our imports.

from pyauthd_serv import BaseMixin
from constants import *
from NSS_constants import *

# NSS specific stuff.

class NssMixin(BaseMixin):
    """Mixin class for BasePyAuthDClient which handles incoming NSS events. It
    implements appropriate handling of NSS structures."""

    # Main handling method. This is called by the Clients handle() method.

    def handle_NSS(self):
        # Receive data based on command.
        comm = self.recvValue()
        rv = None

        # Switch on commands.
        try:
            if comm == NSS_GETPWUID:
                rv = self.handleNss_getPwUid(self.recvValue())
            elif comm == NSS_GETPWNAM:
                rv = self.handleNss_getPwName(self.recvData())
            elif comm == NSS_GETPWENT:
                self.handleNss_getPwEnt()
                return
            elif comm == NSS_GETSPUID:
                rv = self.handleNss_getSpUid(self.recvValue())
            elif comm == NSS_GETSPNAM:
                rv = self.handleNss_getSpName(self.recvData())
            elif comm == NSS_GETSPENT:
                self.handleNss_getSpEnt()
                return
            # Implement the following three properly; for now we have no group
            # handling in this module.
            elif comm == NSS_GETGRGID:
                raise KeyError
            elif comm == NSS_GETGRNAM:
                raise KeyError
            elif comm == NSS_GETGRENT:
                self.sendValue(UNAVAIL)
                return

            # We have a return value, send it and return.
            if rv is not None:
                self.sendValue(SUCCESS)
                rv.sendData(self)
                return

            # We don't have a return value; go to second Exception clause.
            # This means that the current opcode hasn't been worked.
            raise RuntimeError
        except KeyError:
            # User not found error, raised in getUid() or getUser()
            self.sendValue(NOTFOUND)
            self.sendValue(errno.ENOENT)
        except:
            # Other error, we are unavailable, at least if nothing got sent...
            self.sendValue(UNAVAIL)
            if comm not in (NSS_GETPWENT,NSS_GETSPENT,NSS_GETGRENT):
                self.sendValue(errno.ENOENT)

    # NSS methods.

    def handleNss_getPwUid(self,uid):
        return self.getUid(uid).getPasswd()

    def handleNss_getPwName(self,name):
        return self.getUser(name).getPasswd()

    def handleNss_getPwEnt(self):
        self.sendValue(SUCCESS)
        for user in self.listUsers():
            self.sendValue(SUCCESS)
            user.getPasswd().sendData(self)
        self.sendValue(NOTFOUND)

    def handleNss_getSpUid(self,uid):
        return self.getUid(uid).getShadow()

    def handleNss_getSpName(self,name):
        return self.getUser(name).getShadow()

    def handleNss_getSpEnt(self):
        self.sendValue(SUCCESS)
        for user in self.listUsers():
            self.sendValue(SUCCESS)
            user.getShadow().sendData(self)
        self.sendValue(NOTFOUND)

class Passwd(object):

    def __init__(self,name,pwd,uid,gid,gecos,dir,shell):
        self.name = name
        self.pwd = pwd
        self.uid = uid
        self.gid = gid
        self.gecos = gecos
        self.dir = dir
        self.shell = shell

    # Function that uses the client to send the data.

    def sendData(self,client):
        client.sendData(self.name)
        client.sendData(self.pwd)
        client.sendValue(self.uid)
        client.sendValue(self.gid)
        client.sendData(self.gecos)
        client.sendData(self.dir)
        client.sendData(self.shell)

class Shadow(object):

    def __init__(self,name,pwd,lstchg,min,max,warn,inact,expire,flag):
        self.name = name
        self.pwd = pwd
        self.lstchg = lstchg
        self.min = min
        self.max = max
        self.warn = warn
        self.inact = inact
        self.expire = expire
        self.flag = flag

    # Function that uses the client to send the data.

    def sendData(self,client):
        client.sendData(self.name)
        client.sendData(self.pwd)
        client.sendValue(self.lstchg)
        client.sendValue(self.min)
        client.sendValue(self.max)
        client.sendValue(self.warn)
        client.sendValue(self.inact)
        client.sendValue(self.expire)
        client.sendValue(self.flag)
