# -*- coding: iso-8859-15 -*-
"""pyauthd_pam.py

Implementation of the PAM plugin for the python authentication daemon. The PAM
plugin handles all communication with PyAuthD_PAM.c, the PyAuthD PAM module.

Copyright (C) 2005, Heiko Wundram.
Released under a BSD-style license, see LICENSE for details.
"""

# Versioning.
__author__ = "$LastChangedBy: modelnine $"
__version__ = "$LastChangedRevision: 211 $"
__date__ = "$LastChangedDate: 2005-12-15 08:06:01 +0100 (czw, 15 gru 2005) $"

# Python imports.

import bisect
import threading

# Our imports.

from pyauthd_serv import BaseMixin
from constants import *
from PAM_constants import *

# PAM specific stuff.

class PamMixin(BaseMixin):
    """Mixin class for BasePyAuthDClient which handles incoming PAM events. It
    implements appropriate handling for PAM structures."""

    # Main handling methods. These are called by the Clients handle() method.

    def handle_PAM(self):
        # Receive toplevel function arguments.
        funcname = self.recvData()
        flags = self.recvValue()
        argc = self.recvValue()
        argv = [self.recvData() for i in xrange(argc)]
        pamh = PamHandle(self)

        # Try to call function.
        try:
            rv = getattr(self,"handlePam_%s" % funcname)(pamh,flags,argv)
        except:
            rv = SERVICE_ERR

        self.sendValue(PAM_RETURN)
        self.sendValue(rv)

    def handle_PAM_CLEANUP(self):
        # Reveice data handle.
        item = self.recvValue()
        error = self.recvValue()
        pamh = PamHandle(self)

        # Remove data handle from PamHandle class static vars.
        PamHandle.__DATALOCK__.acquire()
        try:
            try:
                data, cleanup = PamHandle.__DATA__[item]
                del PamHandle.__DATA__[item]
                bisect.insort(PamHandle.__DATAITEMS__,item)
                if cleanup is not None:
                    cleanup(pamh,data,error)
            except:
                pass
        finally:
            PamHandle.__DATALOCK__.release()

        # Return success to module.
        self.sendValue(PAM_RETURN)
        self.sendValue(0)

    # PAM methods.

    def handlePam_authenticate(self,pamh,flags,argv):
        # Load the user using the PAM passed user name.
        try:
            print "User not there, getting it using getUser()."
            user = self.getUser(pamh.getUser(None))
        except KeyError:
            print "User unknown!"
            return USER_UNKNOWN
        print "Authenticating user:", user.getUsername()

        # Get authentication token.
        print "Requesting authentication token."
        authtok = pamh.getItem(AUTHTOK)
        if authtok is None:
            authtok, rv = pamh.callConv(PROMPT_ECHO_OFF,"Password:")
            if rv == SUCCESS:
                pamh.setItem(AUTHTOK,authtok)
        if authtok is None:
            print "Getting authentication token failed..."

        # Check authentication token.
        if not user.checkPass(authtok):
            print "Password proved to be wrong."
            return AUTH_ERR

        # Pass authentication.
        print "Authentication successful!"
        return SUCCESS

    def handlePam_setcred(self,pamh,flags,argv):
        return SUCCESS

    def handlePam_acct_mgmt(self,pamh,flags,argv):
        return SUCCESS

    def handlePam_open_session(self,pamh,flags,argv):
        self.getUser(pamh.getUser(None)).openSession()
        return SUCCESS

    def handlePam_close_session(self,pamh,flags,argv):
        self.getUser(pamh.getUser(None)).closeSession()
        return SUCCESS

    def handlePam_chauthtok(self,pamh,flags,argv):
        # Get user.
        user = self.getUser(pamh.getUser(None))

        # Preliminary check is for old password.
        if flags & PRELIM_CHECK:
            # Do we need to check old password? Only if not root (<>0).
            if self.rem_uid:
                # Get old and new authentication token.
                oldauthtok = pamh.getItem(OLDAUTHTOK)
                if oldauthtok is None:
                    oldauthtok, rv = \
                                pamh.callConv(PROMPT_ECHO_OFF,"Old Password:")
                    if rv == SUCCESS:
                        pamh.setItem(OLDAUTHTOK,oldauthtok)

                # Check old authentication token.
                if not user.checkPass(oldauthtok):
                    return TRY_AGAIN

            return SUCCESS

        # Get new token.
        authtok = pamh.getItem(AUTHTOK)
        if authtok is None:
            # Loop until the two authentication tokens match.
            while True:
                # Load once.
                authtok, rv = pamh.callConv(PROMPT_ECHO_OFF,"New Password:")
                if rv <> SUCCESS:
                    return AUTHTOK_ERR

                # Load again.
                authtok_again, rv = \
                               pamh.callConv(PROMPT_ECHO_OFF,
                                             "Retype new Password:")
                if rv <> SUCCESS:
                    return AUTHTOK_ERR

                # Check whether the two match.
                if authtok == authtok_again:
                    break

            # Store authentication token.
            pamh.setItem(AUTHTOK,authtok)

        # Set new password in database.
        user.setPass(authtok)

        return SUCCESS

class PamError(Exception):
    """Base class for all PAM errors."""

    pass

class PamHandle(object):
    """Implements the callback interface to the remote PAM module. The PAM
    handle will contain a reference to the PyAuthDclient instance."""

    __DATA__ = {}
    __DATAITEMS__ = [0]
    __DATALOCK__ = threading.RLock()

    def __init__(self,client):
        self.client = client

    def setData(self,modname,data,cleanup=None):
        # Handle data and store in local data item storage.
        if data is not None:
            self.__DATALOCK__.acquire()
            try:
                newitem = self.__DATAITEMS__.pop(0)
                if not self.__DATAITEMS__:
                    self.__DATAITEMS__.append(newitem+1)
                    self.__DATA__[newitem] = (data,cleanup)
            finally:
                self.__DATALOCK__.release()
        else:
            newitem = -1

        # Set data item to remote host.
        self.client.sendValue(PAM_SET_DATA)
        self.client.sendData(modname)
        self.client.sendValue(newitem)

        # Get return value.
        rv = self.client.recvValue()
        if rv <> SUCCESS:
            raise PamError, rv

    def getData(self,modname):
        # Retrieve data handle from remote.
        self.client.sendValue(PAM_GET_DATA)
        self.client.sendData(modname)

        # Get return value.
        rv = self.client.recvValue()
        if rv <> SUCCESS:
            raise PamError, rv

        # Retrieve item.
        item = self.client.recvValue()

        # Return item.
        self.__DATALOCK__.acquire()
        try:
            return self.__DATA__[item][0]
        finally:
            self.__DATALOCK__.release()

    def setItem(self,item,data):
        # Send item constant and data.
        self.client.sendValue(PAM_SET_ITEM)
        self.client.sendValue(item)
        self.client.sendData(data)

        # Get return value.
        rv = self.client.recvValue()
        if rv <> SUCCESS:
            raise PamError, rv

    def getItem(self,item):
        # Get item.
        self.client.sendValue(PAM_GET_ITEM)
        self.client.sendValue(item)

        # Get return value.
        rv = self.client.recvValue()
        if rv <> SUCCESS:
            raise PamError, rv

        # Get data.
        return self.client.recvData()

    def getUser(self,data):
        # Get user.
        self.client.sendValue(PAM_GET_USER)
        self.client.sendData(data)

        # Get return value.
        rv = self.client.recvValue()
        if rv <> SUCCESS:
            raise PamError, rv

        # Get user.
        return self.client.recvData()

    def putEnv(self,arg,value=None):
        # Construct environment.
        if isinstance(value,str):
            env = "%s=%s" % (arg,value)
        elif value is None:
            env = arg
        else:
            raise TypeError, "Invalid value for variable."

        # Set environment.
        self.client.sendValue(PAM_PUTENV)
        self.client.sendData(env)

        # Get return value.
        rv = self.client.recvValue()
        if rv <> SUCCESS:
            raise PamError, rv

    def getEnv(self,arg):
        # Get environment.
        self.client.sendValue(PAM_GETENV)
        self.client.sendData(arg)

        # Get return value.
        return self.client.recvData()

    def getEnvList(self):
        # Get environment list.
        self.client.sendValue(PAM_GETENVLIST)

        # Retrieve data items.
        envlist = []
        while True:
            item = self.client.recvData()
            if item is None:
                break
            envlist.append(item)

        # Return environment list.
        return envlist

    def getError(self,item):
        # Get error value.
        self.client.sendValue(PAM_GET_ERROR)
        self.client.sendValue(item)

        # Return error value.
        return self.client.recvData()

    def callConv(self,msg_style,msg):
        # Call conversation function.
        self.client.sendValue(PAM_CALL_CONV)
        self.client.sendValue(msg_style)
        self.client.sendData(msg)

        # Retrieve return values (pam_get_item and conv call).
        for i in xrange(2):
            rv = self.client.recvValue()
            if rv <> SUCCESS:
                raise PamError, rv

        # Get return value and response.
        return (self.client.recvData(),self.client.recvValue())

    def failDelay(self,delay):
        # Call faildelay.
        self.client.sendValue(PAM_FAIL_DELAY)
        self.client.sendValue(delay)

        # Retrieve return value.
        rv = self.client.recvValue()
        if rv <> SUCCESS:
            raise PamError, rv
