# -*- coding: iso-8859-15 -*-
"""pyauthd.py

Implements the daemon process which handles PyAuthD contacts. This is a Unix
Domain Socket server which currently is simply based on ThreadingMixIn of the
Python standard UnixStreamServer. This makes the current implementation
vulnerable for DOS attacks. This will be adressed some time in the future.

Copyright (C) 2005, Heiko Wundram.
Released under a BSD-style license, see LICENSE for details.
"""

# Versioning.
__author__ = "$LastChangedBy: modelnine $"
__version__ = "$LastChangedRevision: 78 $"
__date__ = "$LastChangedDate: 2005-10-24 02:41:58 +0200 (pon, 24 paź 2005) $"

# Python modules required by the Python Authentication Daemon server.

from eunuchs import recvmsg
import socket
import SocketServer
import struct

# Our modules required by the Python Authentication Daemon server.

import constants

# Internal constants which aren't exported by Python.

SO_PASSCRED = 16
SCM_CREDENTIALS = 2

# Server class. This is nothing but a UnixStreamServer with ThreadingMixIn.

class PyAuthDServer(SocketServer.ThreadingMixIn,SocketServer.UnixStreamServer):
    """Server for the python authentication daemon. This is basically a
    threaded Unix stream socket server. It enables special handling of the
    socket to enable reception of meta-data."""

    # Initialization of a socket server.

    def __init__(self,*args,**kwargs):
        SocketServer.TCPServer.__init__(self,*args,**kwargs)

        # Initialize quit state.
        self.quit = False

    # Request verification. This just assures that the socket has proper rights
    # to receive credentials.

    def verify_request(self,req,addr):
        # Make sure that we get SCM_CREDENTIALS ancilliary messages.
        req.setsockopt(socket.SOL_SOCKET,SO_PASSCRED,1)

        return True

    # Quit methods.

    def serve_until_quit(self):
        # Handle requests until quit is true.
        while not self.quit:
            self.handle_request()

    def quit(self):
        self.quit = True

# Client classes. A metaclass ensures that clients get a proper handlemap
# which is used by handle() to dispatch appropriately. Actual functionality
# is added in the form of mixin classes available separately.

class _PyAuthDClientMeta(type):
    """Metaclass that transparently handles mixin handlers by creating a
    handle map in a class that contains all dispatchers on first byte which
    are available as handle_* methods. This class should never need to be
    called directly."""

    def __init__(selfcls,name,bases,dct):
        # Copy handlemap of bases.
        handlemap = {}
        for base in reversed(bases):
            try:
                handlemap.update(base.__dict__["__handlemap__"])
            except KeyError:
                pass

        # Fill handlemap with own handlers.
        for name, value in dct.iteritems():
            if name.startswith("handle_") and callable(value):
                try:
                    handlemap[getattr(constants,name[7:])] = value
                except AttributeError:
                    pass

        # Set handlemap for current class.
        setattr(selfcls,"__handlemap__",handlemap)

class BasePyAuthDClient(SocketServer.StreamRequestHandler,object):
    """Implementation of the python authentication daemon client. The client
    handles opcodes which come from the remote end and dispatches
    appropriately."""

    __metaclass__ = _PyAuthDClientMeta

    # Appropriate constants for sending/recieving data.

    INT_TYPE = "=i"
    INT_TYPE_LEN = struct.calcsize(INT_TYPE)

    # Data sending and receiving methods over the socket.

    def recvHandler(self):
        """Receive handler information from the socket and make sure we store
        credentials."""

        try:
            # Receive the handler information and get SCM_CREDENTIALS.
            data, _, _, msgs = recvmsg.recvmsg(self.request.fileno(),flags=0,
                                               maxsize=self.INT_TYPE_LEN)

            # Initialize credentials.
            self.rem_pid, self.rem_uid, self.rem_gid = None, None, None
            for msg in msgs:
                if msg[0] == socket.SOL_SOCKET and msg[1] == SCM_CREDENTIALS:
                    # We found a SCM_CREDENTIALS message, decode it.
                    self.rem_pid, self.rem_uid, self.rem_gid = \
                                  struct.unpack("iii",msg[2])

            # Now, return decoded data.
            return struct.unpack(self.INT_TYPE,data)[0]
        except:
            raise PyAuthDClientError, "Connection dropped unexpectedly."

    def recvValue(self):
        """Recieve integer from socket."""

        try:
            return struct.unpack(self.INT_TYPE,
                                 self.rfile.read(self.INT_TYPE_LEN))[0]
        except:
            raise PyAuthDClientError, "Connection dropped unexpectedly."

    def recvData(self):
        """Recieve string from socket."""

        try:
            datalen = struct.unpack(self.INT_TYPE,
                                    self.rfile.read(self.INT_TYPE_LEN))[0]
            if datalen == -1:
                return None
            return self.rfile.read(datalen)
        except:
            raise PyAuthDClientError, "Connection dropped unexpectedly."

    def sendValue(self,value):
        """Send integer to socket."""

        try:
            self.wfile.write(struct.pack(self.INT_TYPE,value))
        except:
            raise PyAuthDClientError, "Connection dropped unexpectedly."

    def sendData(self,data):
        """Send string to socket."""

        try:
            if data is None:
                self.wfile.write(struct.pack(self.INT_TYPE,-1))
            else:
                self.wfile.write(struct.pack(self.INT_TYPE,len(data)))
                self.wfile.write(data)
        except:
            raise PyAuthDClientError, "Connection dropped unexpectedly."

    # Main dispatcher.

    def handle(self):
        """Handler function which dispatches to the appropriate mixin."""

        # Load communication byte from socket.
        val = self.recvHandler()
        self.__handlemap__[val](self)

    def handle_SERVER_QUIT(self):
        """Handler function for Client which terminates the server. This
        calls into a function of the server to quit the server process on the
        next connect."""

        self.server.quit()

class BaseMixin(object):
    """Base class for all mixins. In case you implement a mixin yourself,
    derive your mixin from this class. This class offers no functionality."""

    __metaclass__ = _PyAuthDClientMeta
