/* PyAuthD_PAM.c
   -------------

   PAM module which connects to a Python script via a Unix Domain Socket. It
   sends all data over the socket, and awaits a return value. The return
   value specifies which PAM method to call on the pam handle.

   The module has no output, but logs critical error conditions to syslog.

   Copyright (C) 2005, Heiko Wundram.
   Released under a BSD-style license, see LICENSE for details.

$LastChangedBy: modelnine $
$LastChangedRevision: 75 $
$LastChangedDate: 2005-10-24 01:51:59 +0200 (pon, 24 paź 2005) $
*/

#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <unistd.h>

/* Make sure all PAM types are defined. */

#define PAM_SM_AUTH
#define PAM_SM_ACCOUNT
#define PAM_SM_SESSION
#define PAM_SM_PASSWORD
#include <security/pam_modules.h>

/* Load specifications common for PyAuthD and PyAuthD_PAM. */

#include "PyAuthD_client.h"
#include "PyAuthD_PAM.h"

int sendFunctionArguments(int sock, const char *funcname, int flags, int argc,
			  const char **argv)
{
  int funcnamelen, argvlen, i;

  SEND_DATA(funcname,funcnamelen);
  SEND_VALUE(flags);
  SEND_VALUE(argc);
  for( i = 0; i < argc; i++ )
    SEND_DATA(argv[i],argvlen);

  return 0;

 SEND_ERROR:
  syslog(LOG_AUTHPRIV|LOG_CRIT,"Unable to send data over socket.");
  return -1;
}

int communicatePam(int sock, pam_handle_t *pamh)
{
  int func, rv, itemtype, datalen, errnum, i, delay;
  void *data, *modname, *moddata, *pconv, *cleanup;
  char *user, *env, **envlist;
  struct pam_conv *conv;
  struct pam_message msg, *pmsg;
  struct pam_response *resp;

  while( 1 ) {
    RECV_VALUE(func);

    switch( func ) {

    case AUTHD_PAM_RETURN:
      /* We should simply return the value that the remote end wants us to. */
      RECV_VALUE(rv);
      return rv;

    case AUTHD_PAM_SET_DATA:
      /* Set module data. */
      RECV_VALUE(datalen);
      RECV_DATA(modname,datalen);
      moddata = malloc(sizeof(int));
      cleanup = &cleanupPam;
      RECV_VALUE((*(int*)moddata));
      if( *((int*)moddata) == -1 ) {
	FREE_DATA(moddata,sizeof(int));
	moddata = NULL;
	cleanup = NULL;
      }
      rv = pam_set_data(pamh,(const char*)modname,moddata,cleanup);
      FREE_DATA(modname,datalen);
      SEND_VALUE(rv);
      break;

    case AUTHD_PAM_GET_DATA:
      /* Retrieve module data. */
      RECV_VALUE(datalen);
      RECV_DATA(modname,datalen);
      rv = pam_get_data(pamh,(const char*)modname,(const void**)&moddata);
      FREE_DATA(modname,datalen);
      SEND_VALUE(rv);
      if( rv == PAM_SUCCESS )
	SEND_VALUE((*(int*)moddata));
      break;

    case AUTHD_PAM_SET_ITEM:
      /* We are requested to set an item. */
      RECV_VALUE(itemtype);
      RECV_VALUE(datalen);
      if( datalen == -1 )
	data = NULL;
      else
	RECV_DATA(data,datalen);
      rv = pam_set_item(pamh,itemtype,data);
      if( datalen != -1 )
	FREE_DATA(data,datalen);
      SEND_VALUE(rv);
      break;

    case AUTHD_PAM_GET_ITEM:
      /* We are requested to get an item. */
      RECV_VALUE(itemtype);
      rv = pam_get_item(pamh,itemtype,(const void**)&data);
      SEND_VALUE(rv);
      if( rv == PAM_SUCCESS ) {
	if( !data ) {
	  datalen = -1;
	  SEND_VALUE(datalen);
	} else
	  SEND_DATA(data,datalen);
      }
      break;

    case AUTHD_PAM_GET_USER:
      /* We are asking for a user to use to authenticate. */
      RECV_VALUE(datalen);
      if( datalen == -1 )
	data = NULL;
      else
	RECV_DATA(data,datalen);
      rv = pam_get_user(pamh,(const char**)&user,(char*)data);
      if( datalen != -1 )
	FREE_DATA(data,datalen);
      SEND_VALUE(rv);
      if( rv == PAM_SUCCESS )
	SEND_DATA(user,datalen);
      break;

    case AUTHD_PAM_PUTENV:
      /* Inserts an environment variable into the pam stack. */
      RECV_VALUE(datalen);
      RECV_DATA(data,datalen);
      rv = pam_putenv(pamh,(const char*)data);
      FREE_DATA(data,datalen);
      SEND_VALUE(rv);
      break;

    case AUTHD_PAM_GETENV:
      /* Requests an environment variable from the pam environment list. */
      RECV_VALUE(datalen);
      RECV_DATA(data,datalen);
      env = (char*)pam_getenv(pamh,(char*)data);
      FREE_DATA(data,datalen);
      if( !env ) {
	datalen = -1;
	SEND_VALUE(datalen);
      } else
	SEND_DATA(data,datalen);
      break;

    case AUTHD_PAM_GETENVLIST:
      /* Acquire the environment list. */
      envlist = (char**)pam_getenvlist(pamh);
      i = 0;
      while( envlist[i] ) {
	SEND_DATA(envlist[i],datalen);
	i += 1;
      }
      datalen = -1;
      SEND_VALUE(datalen);
      break;

    case AUTHD_PAM_GET_ERROR:
      /* We are requested to get the error string for an errnum. */
      RECV_VALUE(errnum);
      data = (void*)pam_strerror(pamh,errnum);
      SEND_DATA(data,datalen);
      break;

    case AUTHD_PAM_CALL_CONV:
      /* We want to call the conversation function. This call only allows a
	 single structure. */
      RECV_VALUE(msg.msg_style);
      RECV_VALUE(datalen);
      RECV_DATA(msg.msg,datalen);
      pmsg = &msg;
      rv = pam_get_item(pamh,PAM_CONV,(const void **)&pconv);
      conv = (struct pam_conv *)pconv;
      SEND_VALUE(rv);
      if( rv != PAM_SUCCESS ) {
	FREE_DATA(((void*)msg.msg),datalen);
	break;
      }
      rv = (*(conv->conv))(1,(const struct pam_message **)&pmsg,&resp,
			   conv->appdata_ptr);
      FREE_DATA(((void*)msg.msg),datalen);
      SEND_VALUE(rv);
      if( rv != PAM_SUCCESS )
	break;
      if( !resp ) {
	datalen = -1;
	SEND_VALUE(datalen);
	SEND_VALUE(rv);
      } else {
	if( !resp->resp ) {
	  datalen = -1;
	  SEND_VALUE(datalen);
	} else {
	  SEND_DATA(resp->resp,datalen);
	  FREE_DATA(resp->resp,datalen);
	}
	SEND_VALUE(resp->resp_retcode);
	FREE_DATA(resp,sizeof(struct pam_response));
      }
      break;

    case AUTHD_PAM_FAIL_DELAY:
      RECV_VALUE(delay);
      rv = pam_fail_delay(pamh,delay);
      SEND_VALUE(rv);
      break;

    default:
      /* Invalid value received from remote end. Signal error. */
      syslog(LOG_AUTHPRIV|LOG_CRIT,"Invalid status code received.");
      goto RECV_ERROR;
    }
  }

 SEND_ERROR:
  syslog(LOG_AUTHPRIV|LOG_CRIT,"Unable to send return value.");
  return -1;

 MEMORY_ERROR:
  syslog(LOG_AUTHPRIV|LOG_CRIT,"Unable to allocate new memory.");
  return -1;

 RECV_ERROR:
  syslog(LOG_AUTHPRIV|LOG_CRIT,"Error receiving data from remote.");
  return -1;
}

void cleanupPam(pam_handle_t *pamh, void *data, int error_status)
{
  int sock;

  if( ( sock = openPyAuthD(AUTHD_PAM_CLEANUP) ) == -1 )
    goto SOCKET_ERROR;

  SEND_VALUE((*(int*)data));
  SEND_VALUE(error_status);

  if( communicatePam(sock,pamh) == -1 )
    goto SOCKET_ERROR;

 SEND_ERROR:
 SOCKET_ERROR:
  memset(data,0,sizeof(int));
  free(data);
  closePyAuthD(sock);
}

/* Authentication functions. */
DECLARE_FUNCTION(authenticate);
DECLARE_FUNCTION(setcred);

/* Account management functions. */
DECLARE_FUNCTION(acct_mgmt);

/* Session management. */
DECLARE_FUNCTION(open_session);
DECLARE_FUNCTION(close_session);

/* Authentication token changing. */
DECLARE_FUNCTION(chauthtok);
